﻿/* The MIT License (MIT)
 * 
 * Copyright (c) 2010 Michael Horstmann
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.   
 * 
 */

using System;
using System.IO;
using System.Text;

namespace IntelHexParser
{
    public class IntelHexBin
    {
        protected struct HexRecord
        {
            public byte Length;
            public int Offset;
            public byte Type;
            public byte[] Data;
        }

        protected byte[] _memoryMap;

        /// <summary>
        /// start offset within hex memory map
        /// </summary>
        protected int _startOffset;

        /// <summary>
        /// end offset within hex memory map
        /// </summary>
        protected int _endOffset;

        /// <summary>
        /// flag, wich indicates, if a hexfile was already successfully loaded in.
        /// </summary>
        protected bool _hasValidData;

        /// <summary>
        /// Memory map which contains the data of a loaded file (Intel HEX or a binary one)
        /// </summary>
        /// <value>Memory map.</value>
        public byte[] MemoryMap
        {
            get { return _memoryMap; }
        }

        /// <summary>
        /// Gets the size in bytes of the currently used memory in the map
        /// </summary>
        /// <value>The size of the used.</value>
        public int UsedMemory
        {
            get { return _endOffset - _startOffset + 1; }
        }

        /// <summary>
        /// Gets a value indicating whether the parser instance has valid data.
        /// </summary>
        /// <value>
        /// <c>true</c> if the parser has valid data; otherwise, <c>false</c>.
        /// </value>
        public bool HasValidData
        {
            get { return _hasValidData; }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="IntelHexBin" /> class.
        /// </summary>
        /// <param name="bufferSize">Size of the buffer [bytes].</param>
        /// <exception cref="System.ArgumentException">Invalid HEX buffer size!</exception>
        public IntelHexBin(int bufferSize)
        {
            if (bufferSize <= 0) throw new ArgumentException("Invalid HEX buffer size!");

            _memoryMap = new byte[bufferSize];
            ResetMemoryMap();
        }

        /// <summary>
        /// Resets memory map and status.
        /// </summary>
        public void ResetMemoryMap(byte value = 0x00)
        {
            for (var i = 0; i < _memoryMap.Length; i++)
                _memoryMap[i] = value;

            _startOffset = 0;
            _endOffset = 0;
            _hasValidData = false;
        }

        /// <summary>
        /// Convert a hexadecimal value to decimal value.
        /// </summary>
        /// <param name="hexString">A hexadecimal string.</param>
        private static int Hex2Dec(string hexString)
        {
            var value = 0;
            if (hexString.Length == 0) throw new ArgumentException("Unable to convert zero-length hex string to number!");
            if (hexString.Length > 8) throw new ArgumentOutOfRangeException("hexString");

            for (var i = 0; i < hexString.Length; i++)
            {
                var hexDigit = hexString.Substring(i, 1);
                value = (value << 4) | Convert.ToByte(hexDigit, 16);
            }
            return value;
        }

        /// <summary>
        /// Converts a decimal value to a hexadecimal string.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <param name="minWidth">Width of the min.</param>
        private static string Dec2Hex(int value, int minWidth)
        {
            if (minWidth < 0 || minWidth > 8) throw new ArgumentOutOfRangeException("minWidth");
            return string.Format("{0:X"+ minWidth + "}", value);
        }

        /// <summary>
        /// Parses one record.
        /// </summary>
        /// <param name="hexLine">Current dataset line.</param>
        /// <param name="record">Record reference</param>
        protected static EIntelHexResult ParseRecord(string hexLine, ref HexRecord record)
        {
            if (hexLine.Length < 11)
                return EIntelHexResult.MissingFieldsInLine;

            // we accept only "clean" hexfiles, without any characters before the colon!
            if (!hexLine.StartsWith(":"))
                return EIntelHexResult.ColonExpected;

            try
            {
                record.Length = Convert.ToByte(Hex2Dec(hexLine.Substring(1, 2)));
                record.Offset = Hex2Dec(hexLine.Substring(3, 4));
                record.Type = Convert.ToByte(Hex2Dec(hexLine.Substring(7, 2)));
            }
            catch
            {
                return EIntelHexResult.HexConversionFailed;
            }

            if (hexLine.Length < 11 + record.Length * 2)
                return EIntelHexResult.MissingFieldsInLine;

            // process checksum
            var checksum = record.Length;
            checksum += Convert.ToByte((record.Offset >> 8) & 0xFF);
            checksum += Convert.ToByte(record.Offset & 0xFF);
            checksum += record.Type;

            if (record.Length > 0)
            {
                record.Data = new byte[record.Length];

                for (var recordPos = 0; recordPos < record.Length; recordPos++)
                {
                    record.Data[recordPos] = Convert.ToByte(Hex2Dec(hexLine.Substring(9 + Convert.ToInt16(recordPos * 2), 2)));
                    // add to 256-mod checksum
                    checksum+=record.Data[recordPos];
                }
            }
            checksum += Convert.ToByte(Hex2Dec(hexLine.Substring(9 + Convert.ToInt16(record.Length * 2), 2)));

            return checksum != 0 ? EIntelHexResult.ChecksumError : EIntelHexResult.Success;
        }

        /// <summary>
        /// Writes a HEX record to overgiven stream writer
        /// </summary>
        /// <param name="streamWriter">The sw.</param>
        /// <param name="record">The rec.</param>
        protected static void WriteRecord(ref StreamWriter streamWriter, ref HexRecord record)
        {
            int recordPos;
            var checksum = record.Length;
            checksum += Convert.ToByte((record.Offset >> 8) & 0xFF);
            checksum += Convert.ToByte(record.Offset & 0xFF);
            checksum += record.Type;

            // write initial line
            streamWriter.Write(":" + Dec2Hex(record.Length, 2) + Dec2Hex(record.Offset, 4) + Dec2Hex(record.Type, 2));

            for (recordPos = 0; recordPos < record.Length; recordPos++)
            {
                checksum += record.Data[recordPos];
                streamWriter.Write(Dec2Hex(record.Data[recordPos], 2));
            }

            // make two's complement of 256-modulo-checksum
            checksum = (byte)(checksum != 0 ? Convert.ToByte(256 - checksum) : 0);

            // and finally write the hexline
            streamWriter.WriteLine(Dec2Hex(checksum, 2));
        }

        /// <summary>
        /// Loads an internally stream which references to a Intel HEX file.
        /// </summary>
        /// <param name="hexStream">The stream from an Intel HEX file.</param>
        protected EIntelHexResult LoadHexStream(StreamReader hexStream)
        {
            if (hexStream == null) throw new ArgumentNullException("hexStream");

            var rec = new HexRecord();
            var baseAddress = 0;

            // first reset current memory map
            ResetMemoryMap();
            _startOffset = Convert.ToInt32(_memoryMap.Length);

            while (!hexStream.EndOfStream)
            {
                string hexline = hexStream.ReadLine();
                var parseResult = ParseRecord(hexline, ref rec);

                if (parseResult != EIntelHexResult.Success)
                {
                    hexStream.Close();
                    return parseResult;
                }

                switch (rec.Type)
                {
                    // DATA RECORD
                    // so copy the data
                    case 0x00:

                        if (baseAddress + rec.Offset + rec.Length > _memoryMap.Length)
                        {
                            hexStream.Close();
                            return EIntelHexResult.DataOutsideMemoryMapDefined;
                        }

                        for (var dataPos = 0; dataPos < rec.Length; dataPos++)
                            _memoryMap[baseAddress + rec.Offset + dataPos] = rec.Data[dataPos];

                        // update byte usage
                        if (baseAddress + rec.Offset < _startOffset)
                            _startOffset = baseAddress + rec.Offset;

                        if (baseAddress + rec.Offset + rec.Length - 1 > _endOffset)
                            _endOffset = baseAddress + rec.Offset + rec.Length - 1;
                        break;

                    // EXTENDED SEGMENT ADDRESS RECORD
                    case 0x02:
                        baseAddress = Convert.ToInt32(rec.Data[0] << 8) | rec.Data[1];
                        baseAddress <<= 4;
                        break;

                    // START SEGMENT ADDRESS RECORD
                    // address field contains 0 and data exists at entry point for program execution as CS:IP on x86 CPUs
                    case 0x03:
                        break;

                    // EXTENDED LINEAR ADDRESS RECORD
                    // address field contains 0 and data contains the high word a 32bit linear adress record,
                    // whose following data hex lines records are to be stored
                    case 0x04:
                        baseAddress = Convert.ToInt32(rec.Data[0] << 8) | rec.Data[1];
                        baseAddress <<= 16;
                        break;

                    // START LINEAR ADDRESS RECORD
                    // the address field contains 0 and data is the entry point for program execution as EIP
                    case 0x05:
                        break;

                    // END OF FILE file record (EOF)
                    case 0x01:
                        hexStream.Close();
                        return EIntelHexResult.Success;

                    // INVALID TYPE RECEIVED
                    default:
                        hexStream.Close();
                        return EIntelHexResult.UnknownHexFormat;
                }
            }

            hexStream.Close();
            return EIntelHexResult.UnknownHexFormat;
        }

        /// <summary>
        /// Loads the hex file.
        /// </summary>
        /// <param name="path">The path.</param>
        public virtual EIntelHexResult LoadHexFile(string path)
        {
            try
            {
                using (var readFile = new StreamReader(path))
                {
                    // now try to load in the hex stream
                    var loadResult = LoadHexStream(readFile);

                    if (loadResult == EIntelHexResult.Success)
                        _hasValidData = true;

                    return loadResult;
                }
            }
            catch
            {
                return EIntelHexResult.FileError;
            }
        }

        /// <summary>
        /// Loads from an existing stream which references to a Intel HEX file.
        /// </summary>
        /// <param name="stream">The stream.</param>
        public virtual EIntelHexResult LoadHexFile(Stream stream)
        {
            try
            {
                using (var readFile = new StreamReader(stream))
                {
                    // now try to load in the hex stream
                    var loadResult = LoadHexStream(readFile);

                    if (loadResult == EIntelHexResult.Success)
                        _hasValidData = true;

                    return loadResult;
                }
            }
            catch
            {
                return EIntelHexResult.FileError;
            }
        }

        /// <summary>
        /// Saves a Intel HEX file from existing memory map.
        /// </summary>
        /// <param name="path">The file path.</param>
        public virtual EIntelHexResult SaveHexFile(string path)
        {
            var hexfile = new StreamWriter(path, false, Encoding.ASCII);

            try
            {
                var rec = new HexRecord
                {
                    Data = new byte[16]
                };
                var baseAddress = _startOffset;
                baseAddress &= ~Convert.ToInt32(0xFFFF);
                var offset = Convert.ToInt32(_startOffset & 0xFFFF);
                var dataPos = 0;
                rec.Length = 2;
                rec.Offset = 0;
                rec.Type = 0x02;
                rec.Data[1] = 0x00;
                rec.Data[0] = Convert.ToByte((baseAddress >> 12));

                // write initial record
                WriteRecord(ref hexfile, ref rec);

                do
                {
                    // check for passed 64k boundaries
                    if (offset + dataPos >= 0x10000)
                    {
                        offset -= 0x10000;
                        baseAddress += 0x10000;
                        rec.Length = 2;
                        rec.Offset = 0;
                        rec.Type = 0x02;
                        // give 4k page index
                        rec.Data[0] = Convert.ToByte(baseAddress >> 12);
                        rec.Data[1] = 0x00;
                        WriteRecord(ref hexfile, ref rec);
                    }

                    // put data into record
                    rec.Data[dataPos] = _memoryMap[baseAddress + offset + dataPos];
                    dataPos++;

                    // Check if we need to write out the current data record
                    if (offset + dataPos > 0x10000 || dataPos >= 16 || baseAddress + offset + dataPos > _endOffset)
                    {
                        // write current data record
                        rec.Length = Convert.ToByte(dataPos);
                        rec.Offset = offset;
                        rec.Type = 0x00;

                        WriteRecord(ref hexfile, ref rec);
                        offset += dataPos;
                        dataPos = 0;
                    }

                } while (baseAddress + offset + dataPos <= _endOffset);

                // finally write EOF record
                rec.Length = 0;
                rec.Offset = 0;
                rec.Type = 0x01;
                WriteRecord(ref hexfile, ref rec);

                hexfile.Close();
                hexfile.Dispose();
                return EIntelHexResult.Success;
            }
            catch
            {
                return EIntelHexResult.UnknownHexFormat;
            }
            finally
            {
                if (hexfile != null)
                    hexfile.Dispose();
            }
        }

        /// <summary>
        /// Loads a binary stream.
        /// </summary>
        protected EIntelHexResult LoadBinStream(BinaryReader binaryReader)
        {
            try
            {
                ResetMemoryMap();
                _memoryMap = binaryReader.ReadBytes(Convert.ToInt32(binaryReader.BaseStream.Length));

                // we don't know any offset when we load simple binary files, so simply setup the offsets based from beginning at address 0
                _endOffset = Convert.ToInt32(binaryReader.BaseStream.Length - 1);
                _hasValidData = true;
            }
            catch
            {
                return EIntelHexResult.DataOutsideMemoryMapDefined;
            }
            return EIntelHexResult.Success;
        }

        /// <summary>
        /// Loads a binary file from an overgiven filepath.
        /// </summary>
        /// <param name="path">The path.</param>
        public virtual EIntelHexResult LoadBinFile(string path)
        {
            try
            {
                using (var readFile = new FileStream(path, FileMode.Open, FileAccess.Read))
                {
                    if (readFile.Length > _memoryMap.Length)
                        return EIntelHexResult.BinaryFileIsTooBig;

                    var binReader = new BinaryReader(readFile);
                    var loadResult = LoadBinStream(binReader);

                    if (loadResult == EIntelHexResult.Success)
                        _hasValidData = true;

                    return loadResult;
                }
            }
            catch
            {
                return EIntelHexResult.FileError;
            }
        }

        /// <summary>
        /// Loads an existing binary file as an overgiven stream.
        /// </summary>
        /// <param name="stream">The stream.</param>
        public virtual EIntelHexResult LoadBinFile(Stream stream)
        {
            if (stream.Length > _memoryMap.Length)
                return EIntelHexResult.BinaryFileIsTooBig;

            try
            {
                using (var readFile = new BinaryReader(stream))
                {
                    var loadResult = LoadBinStream(readFile);

                    if (loadResult == EIntelHexResult.Success)
                        _hasValidData = true;

                    return loadResult;
                }
            }
            catch
            {
                return EIntelHexResult.FileError;
            }
        }

        /// <summary>
        /// Saves a binary file to an overgiven path.
        /// </summary>
        /// <param name="path">The path.</param>
        /// <param name="writeUnusedMemory">if set to <c>true</c> [write unused memory].</param>
        public virtual EIntelHexResult SaveBinFile(string path, bool writeUnusedMemory)
        {
            try
            {
                using (var saveFile = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.None))
                {
                    var binWriter = new BinaryWriter(saveFile);

                    if (saveFile.Length > _memoryMap.Length)
                        return EIntelHexResult.BinaryFileIsTooBig;

                    if (writeUnusedMemory)
                        binWriter.Write(MemoryMap);
                    else
                        binWriter.Write(MemoryMap, Convert.ToInt32(_startOffset), Convert.ToInt32(_endOffset) + 1);

                    return EIntelHexResult.Success;
                }
            }
            catch
            {
                return EIntelHexResult.BinaryExceptionError;
            }
        }
    }
}