﻿/* The MIT License (MIT)
 * 
 * Copyright (c) 2010 Michael Horstmann
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.   
 * 
 */
using System;
using System.IO;

namespace IntelHexParser
{
    internal class Program
    {
        private static IntelHexBin _intelHex;

        private static void Main(string[] args)
        {
            string filePath;

            try
            {
                // first initialize Intel HEX parser with 10MB of static memory
                _intelHex = new IntelHexBin(1024 * 1024 * 10);
            }
            catch (Exception)
            {
                Console.WriteLine("Error occured during initialization of Intel HEX file parser!");
                return;
            }

            if (args.Length > 0)
            {
                filePath = args[0];
                if (!LoadHexFile(filePath)) return;
            }
            else
            {
                Console.Write("Please specify a Intel HEX file to load (maximum memory to use <= 10 MByte): ");
                filePath = Console.ReadLine();
                if (!LoadHexFile(filePath)) return;
            }

            Console.Write("Save also test files for demonstration (test123.bin + test123.hex) (y/n)? : ");

            if (Console.ReadKey(false).KeyChar == 'y')
            {
                Console.WriteLine(string.Empty);
                SaveDemoFiles(filePath);
            }
        }

        /// <summary>
        /// Loads a given Intel HEX file.
        /// </summary>
        /// <param name="file">The file.</param>
        /// <returns>true, if operation succeeded, otherwise false.</returns>
        private static bool LoadHexFile(string file)
        {
            var result = _intelHex.LoadHexFile(file);

            if (result == EIntelHexResult.Success)
            {
                Console.WriteLine("File loaded successfully!");
                Console.WriteLine("Memory in use (Bytes): {0}", _intelHex.UsedMemory);
                return true;
            }

            Console.WriteLine("Error ({0}) occured when trying to load file '{1}'!", GetErrorCodeString(result), file);
            return false;
        }

        /// <summary>
        /// Saves the demo files.
        /// </summary>
        /// <param name="existingFilePath">The existing file path of the source HEX file.</param>
        private static void SaveDemoFiles(string existingFilePath)
        {
            var path = Path.GetDirectoryName(existingFilePath);
            if (path == null) return;

            Console.Write("Saving binary version for demonstration to test123.bin ...");
            var testBinfile = Path.Combine(path, "test123.bin");
            File.Delete(testBinfile);
            Console.WriteLine(_intelHex.SaveBinFile(testBinfile, true) == EIntelHexResult.Success ? "ok!" : "error!");

            Console.Write("Saving Intel HEX version for demonstration to test123.hex ...");
            var testHexfile = Path.Combine(path, "test123.hex");
            File.Delete(testHexfile);
            Console.WriteLine(_intelHex.SaveHexFile(testHexfile) == EIntelHexResult.Success ? "ok!" : "error!");
        }

        /// <summary>
        /// Gets the error code string.
        /// </summary>
        /// <param name="result">The result code.</param>
        private static string GetErrorCodeString(EIntelHexResult result)
        {
            switch (result)
            {
                case EIntelHexResult.Success:
                    return "Success";

                case EIntelHexResult.MissingFieldsInLine:
                    return "Missing fields in line!";

                case EIntelHexResult.ColonExpected:
                    return "Colon in file expected!";

                case EIntelHexResult.HexConversionFailed:
                    return "Conversion to HEX file falied!";

                case EIntelHexResult.ChecksumError:
                    return "Checksum error!";

                case EIntelHexResult.FileError:
                    return "File error or not found!";

                case EIntelHexResult.UnknownHexFormat:
                    return "Unknown Intel HEX format!";

                case EIntelHexResult.DataOutsideMemoryMapDefined:
                    return "Data outside memory defined!";

                case EIntelHexResult.BinaryFileIsTooBig:
                    return "The binary file is too big for specified memory usage!";

                case EIntelHexResult.BinaryExceptionError:
                    return "Binary exception error!";

                default:
                    return string.Empty;
            }
        }
    }
}
